/* 
 * db.c -- manage a table of query sequences.  
 *
 * The entry point is
 *    int qdb_fill_block(FILE *fp, qdb_t *qdb, int do_revcomp)
 *
 * The input is a seekable file, and 
 * the result is a qdb structure containing the query sequence, its
 * appended reverse complement, and header information describing
 * the beginning of each constituent sequence, its original FASTA
 * header, and other ancillary things.
 */

#ifndef __lint
static const char rcsid[] =
"$Id: db.c,v 1.16 1999/04/28 16:44:42 schwartz Exp $";
#endif

#include "psublast/libc.h"
#include "psublast/psublast.h"
#include "db.h"

enum { Ok, Full, Eof };

void fmt_id(char *h, int c, char* p)
{
    snprintf(h, HDRSAVE, "%c%.*s", c, HDRSAVE-2, p);
}

static void save_id(qdb_t *qdb, int i, int c, char *p)
{
    fmt_id(qdb->hdr[i], c, p);
}

/* read_id -- read the id line from fp, and save some of it. */
static int read_id(FILE *fp, qdb_t *qdb, int i)
{
    char buf[1024]; /* XXX - arbitrary limit */
    int n;

    if (fgets(buf, sizeof(buf), fp) == 0)
        return Eof;
    buf[strlen(buf)-1] = 0;
    n = strspn(buf+1, " \t");
    save_id(qdb, i, '+', buf+1+n);
    return Ok;
}

/* copy sequence data from fp into qdb->p[].
 * return pointer to end, or 0 if overflowed.
 * rely on caller to update qdb->p for next time.
 */
static unsigned char *read_seq(FILE *fp, qdb_t *qdb)
{
    unsigned char *p = qdb->p;

#if 0
    *p++ = 'X';
#endif
    while (p < qdb->sequence+BLOCKLIM) {
        int c = getc(fp);
	/* test most common case first.  
           the manpage says EOF is an acceptable argument for is*(). */
        if (isalpha(c)) {
            *p++ = c & 0xFF;
        } else if (isspace(c)) {
            ; /* skip it */
        } else if (ispunct(c)) {
            ungetc(c, fp);
            return p;
        } else if (c == EOF) {
            return p;
        } else {
            /* can't happen, hopefully */
            fprintf(stderr, "bad character '%c' in read_seq", c);
            exit(1);
        }
    }
    return 0;
}
  
/* read_one --- Read one sequence from fp.
 * Return Ok for success, Full if the qdb table is full,
 * and Eof if we ran out of input.  
 * The file must be seekable.
 */
static int read_one(FILE *fp, qdb_t *qdb, int *len)
{
    unsigned char *q;
    int i = qdb->nseq;

    if (i >= SEQLIM)
        return Full;

    if (feof(fp))
        return Eof;

    qdb->fpos[i] = ftell(fp);
    qdb->beg[i] = qdb->p - qdb->sequence;
    qdb->hdr[i][0] = 0;

    if (i >= SEQLIM)
        return Full;

    read_id(fp, qdb, i);
    if ((q = read_seq(fp, qdb)) != 0) {
	/* got one */
        *q++ = 'X';
	*q = 0;
	*len = q - qdb->p;
	qdb->p = q;
        qdb->beg[i+1] = qdb->p - qdb->sequence;
        return Ok;
    } else {
	/* rewind for next time */
	fseek(fp, qdb->fpos[i], SEEK_SET);
	*qdb->p = 0;
        return Full;
    }
}

/* read_fasta -- read as many as possible sequences from
 * a FASTA format file, and store them in the qdb.
 */
static int read_fasta(FILE *fp, qdb_t *qdb)
{
    int len = 0;
    qdb->nseq = 0;
    qdb->p = qdb->sequence;
    while (read_one(fp, qdb, &len) == Ok)
        ++qdb->nseq;
    return qdb->nseq;
}

/* make_revcomp_seq -- given a query sequence in qdb, construct
 * its reverse complement and append it to the original.
 * assume that there is enough space.
 */
static void make_revcomp_seq(qdb_t *qdb)
{
    unsigned char *p = qdb->p;
    unsigned char *rp = p;
    --p; /* p's nul and preceeding 'X' go at end. */
    while (--p >= qdb->sequence)
        *rp++ = dna_cmpl(*p);
    *rp++ = 'X';
    *rp = 0;
}

/* make_revcomp_tab -- add new header information to a qdb
 * to reflect the recently appended reverse complement.
 * assume that there is enough space.
 */
static void make_revcomp_tab(qdb_t *qdb)
{
    int i, j;
    i = qdb->nseq;
    for (j = 1; j <= i; ++j) {
        qdb->beg[i+j] = qdb->beg[i]+(qdb->beg[i]-qdb->beg[i-j]);
        save_id(qdb, i+j-1, '-', qdb->hdr[i-j]+1);
    }
    qdb->fpos[2*i] = -1;
    qdb->hdr[2*i][0] = 0;
}

/* make_revcomp -- modify qdb in place by appending the query
 * sequence's reverse complement, and updating the headers to match.
 * return the new size of the table.
 */
static int make_revcomp(qdb_t *qdb)
{
    make_revcomp_seq(qdb);
    make_revcomp_tab(qdb);
    qdb->nseq *= 2;
    return qdb->nseq;
}

/* report -- print qdb headers for debugging */
static void report(int i, qdb_t *qdb)
{
    int j;

    printf("i == %d\n", i);
    for (j = 0; j <= i; ++j) {
        printf("i:%d fpos:%ld beg:%ld hdr:'%s'\n",
            j, qdb->fpos[j], qdb->beg[j], qdb->hdr[j]);
    }
    j = strlen((const char*)qdb->sequence);
    printf("len(sequence)=%d\n", j);
    if (j < 1024) printf("%s\n", qdb->sequence);
}

/* qdb_fill_block -- Read a number of sequences from a FASTA
 * format file and store them and their reverse-complements
 * in a query database structure.  The file must be seekable.
 */
int qdb_fill_block(FILE *fp, qdb_t *qdb, int do_revcomp)
{
    int i = read_fasta(fp, qdb);
    if (do_revcomp)
        i = make_revcomp(qdb);
    return i;
}

int qdb_is_revcomp(qdb_t *qdb, int i)
{
    assert(0 <= i);
    assert(i < qdb->nseq);
    return qdb->hdr[i][0] == '-';  /* XXX -- kludge */
}

int qdb_length(const qdb_t *qdb, int i)
{
    assert(0 <= i);
    assert(i < qdb->nseq);
    return qdb->beg[i+1] - qdb->beg[i];
}

int qdb_total_length(const qdb_t *qdb)
{
    return qdb->beg[qdb->nseq];
}

const uchar *qdb_seq(qdb_t *qdb, int i)
{
    assert(0 <= i);
    assert(i < qdb->nseq);
    return qdb->sequence + qdb->beg[i];
}

#ifdef TEST
qdb_t qdb;
int main(int argc, char **argv)
{
    FILE *fp = fopen(argv[1], "r");
    while (!feof(fp)) {
        int i = qdb_fill_block(fp, &qdb);
        report(i, &qdb);
        putchar('\n');
    }
}
#endif
