#include "psublast/libc.h"
#include "psublast/psublast.h"

#ifndef __lint
static const char rcsid[] =
"$Id: blast_table.c,v 1.30 1999/09/23 21:32:35 schwartz Exp schwartz $";
#endif

typedef struct blast_pos_node {
	struct blast_pos_node *link;	/* next occurrence of the word */
	int pos;		/* position where word occurs in seq #1 */
} blast_pos_node_t;

typedef struct blast_epos_node {
	struct blast_epos_node *link;	/* next occurrence of the word */
	int pos;		/* position where word occurs in seq #1 */
	blast_ecode_t ecode;
} blast_epos_node_t;

static void blast_free_chain(struct blast_table *bt, blast_ecode_t ecode);

/* -------------   build table of W-tuples in the first sequence ------------*/

msp_table_t *msp_new_table(void)
{
	msp_table_t *mt = ckallocz(sizeof(*mt));
	return mt;
}

static blast_pos_node_t *cons_pos_node(int pos, blast_ecode_t ecode, blast_pos_node_t *link)
{
	blast_pos_node_t *n = ckalloc(sizeof(blast_pos_node_t));
	n->link = link;
	n->pos = pos;
	ecode = 0; /* XXX - USED(ecode); */
	return n;
}

static blast_epos_node_t *cons_epos_node(int pos, blast_ecode_t ecode, blast_epos_node_t *link)
{
	blast_epos_node_t *n = ckalloc(sizeof(struct blast_epos_node));
	n->link = link;
	n->pos = pos;
	n->ecode = ecode;
	return n;
}

static unsigned int pos_add(blast_table_t *bt, blast_ecode_t ecode, int pos)
{
	unsigned int e = ecode % BLAST_POS_TAB_SIZE;
	bt->u.epos_tab[e] = cons_epos_node(pos, ecode, bt->u.epos_tab[e]);
	return e;
}

static blast_epos_node_t **pos_getp(const blast_table_t *bt, blast_ecode_t ecode)
{
	unsigned int e = ecode % BLAST_POS_TAB_SIZE;
	return &bt->u.epos_tab[e];
}

static blast_epos_node_t *pos_get(const blast_table_t *bt, blast_ecode_t ecode)
{
	return *pos_getp(bt, ecode);
}

static int W_is_big(int W)
{
	return (1U<<(W*2)) > BLAST_POS_TAB_SIZE;
}

static blast_table_t *cons_blast_table(int W)
{
	blast_table_t *bt = ckallocz(sizeof(*bt));
	bt->encoding = fasta_encoding;
	if ((unsigned int)W > sizeof(blast_ecode_t)*8/2)
		W = sizeof(blast_ecode_t)*8/2;
	bt->W = W;
	bt->u.epos_tab=ckallocz(BLAST_POS_TAB_SIZE*sizeof(blast_epos_node_t));
/*
	if (W_is_big(W)) ... else
	bt->u.pos_tab=ckallocz(BLAST_POS_TAB_SIZE*sizeof(blast_pos_node_t));
*/
/*
	pos_pos = ckalloc(tabsize * sizeof(long));
	pos_next = ckalloc(tabsize * sizeof(long));
	pos_ecode = ckalloc(tabsize * sizeof(long));
*/
	return bt;
}

blast_table_t *blast_new_table(SEQ *seq, int W)
{
	blast_ecode_t ecode;
	int i;
	uchar *s;

	uchar *const seq_beg = SEQ_CHARS(seq);
	uchar *const seq_end = seq_beg+SEQ_LEN(seq);
	blast_table_t *const bt = cons_blast_table(W);
	const signed char *const encoding = bt->encoding;
	const blast_ecode_t mask = (1UL << ((W-1)*2)) - 1;

        if (W<=1) return 0;
	for (s = seq_beg; s < seq_end; ) {
	restart:
		/* init */
		ecode = 0L;
		for (i = 1; i < W && s < seq_end; ++i) {
			const int e = encoding[*s++];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ecode = (ecode << 2) + e;
		}
		/* compute */
		for (; s < seq_end; ++s) {
			const int e = encoding[*s];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ecode = ((ecode & mask) << 2) + e;
			pos_add(bt, ecode, s-seq_beg+1);
		} 
	}
	return bt;
}
     
static int chain_len(blast_table_t *bt, blast_ecode_t ecode)
{
        blast_epos_node_t *p;
	int count = 0;
	for (p = pos_get(bt, ecode); p; p = p->link)
		++count;
	return count;
}

static int pos_density(blast_table_t *bt, blast_ecode_t ecode)
{
        blast_epos_node_t *p;
	int count = 0;
	for (p = pos_get(bt, ecode); p; p = p->link)
		++count;
	return count;
}

blast_table_t *blast_table_pare(blast_table_t *bt, int surfeit)
{
        blast_ecode_t ecode;

        if (bt == 0) return 0;
 
        for (ecode = 0; ecode < BLAST_POS_TAB_SIZE; ++ecode) {
		/* Delete ecodes whose chains are too long. */
		if (chain_len(bt, ecode) > surfeit)
			blast_free_chain(bt, ecode);
#if 0
		/* Delete ecodes whose positions are too dense. */
		if (pos_density(bt, ecode) > density)
			blast_free_chain(bt, ecode);
#endif
	}
	return bt;
}

/* -----------------------   search the second sequence   --------------------*/

static msp_table_t *blast_search_unchecked(SEQ *seq1, SEQ *seq2, struct blast_table *bt, ss_t ss, int X, int K)
{
	uchar *s;
	blast_ecode_t ecode;
	int *diag_lev, *dl;
	int i;

	const int W = bt->W;
	const int len1 = SEQ_LEN(seq1);
	/*uchar *const beg1 = SEQ_CHARS(seq1);*/
	/*uchar *const end1 = beg1+len1;*/
	const int len2 = SEQ_LEN(seq2);
	uchar *const beg2 = SEQ_CHARS(seq2);
	uchar *const end2 = beg2+len2;
	const signed char *const encoding = bt->encoding;
	const blast_ecode_t mask = (1UL << ((W-1)*2)) - 1;
	msp_table_t *const msp_tab = msp_new_table();

        if (W<=1) return 0;

	dl = ckallocz((len1+len2+1)*sizeof(int));
	diag_lev = dl + len1;

	for (s = beg2; s < end2; ) {
	restart:
		/* init */
		ecode = 0;
		for (i = 1; i < W && s < end2; ++i) {
			const int e = encoding[*s++];
			if (e < 0) goto restart;
			ecode = (ecode << 2) + e;
		}
		/* compute */
		for (; s < end2; ++s) {
			const int e = encoding[*s];
			const blast_epos_node_t *h;
			if (e < 0) goto restart; /* hit a run of 'X's */
			ecode = ((ecode & mask) << 2) + e;
			for (h = pos_get(bt, ecode); h; h = h->link)
				if (h->ecode == ecode)
					msp_extend_hit(msp_tab, seq1, seq2, 
						ss, X, K, W,
						h->pos, s-beg2+1, diag_lev);
		} 
	}

	ZFREE(dl);
	return msp_tab;
}

msp_table_t *blast_search(SEQ *seq1, SEQ *seq2, struct blast_table *bt, ss_t ss, int X, int K)
{
	if (seq1 == 0) return 0;
	if (seq2 == 0) return 0;
	if (bt == 0) return 0;

	return blast_search_unchecked(seq1, seq2, bt, ss, X, K);
}

/* extend_hit - extend a word-sized hit to a longer match */
int msp_extend_hit(msp_table_t *mt,
			SEQ *s1, SEQ *s2, ss_t ss, int X, int K, int W,
			int pos1, int pos2, int *diag_lev)
{
	const uchar *beg1, *end1;
	int left_sum, right_sum;
	const uchar *const seq1 = SEQ_CHARS(s1);
	const uchar *const seq2 = SEQ_CHARS(s2);
	const int diag = pos2 - pos1;

	if (seq1 == seq2 && pos1 >= pos2)
		return 0;
	
	if (diag_lev[diag] > pos1)
		return 0;
        
	{ /* extend to the right */
	    const uchar *q = end1 = seq1 + pos1;
	    const uchar *s = seq2 + pos2;
            const int end = SEQ_LEN(s2)-diag;
	    const int end_ = (end < SEQ_LEN(s1)) ? end : SEQ_LEN(s1);
	    const uchar *const seq1_end = seq1 + end_;
	    int sum;

#ifdef UNROLL  /* XXX */
	    right_sum = sum = 0;
	    while ((seq1_end - q) > 4) {
		if (sum < right_sum - X) break;
		sum += ss[*s++][*q++];
		if (sum > right_sum) { right_sum = sum; end1 = q; }

		if (sum < right_sum - X) break;
		sum += ss[*s++][*q++];
		if (sum > right_sum) { right_sum = sum; end1 = q; }

		if (sum < right_sum - X) break;
		sum += ss[*s++][*q++];
		if (sum > right_sum) { right_sum = sum; end1 = q; }

		if (sum < right_sum - X) break;
		sum += ss[*s++][*q++];
		if (sum > right_sum) { right_sum = sum; end1 = q; }
	    }
	    while (q < seq1_end) {
		if (sum < right_sum - X) break;
		sum += ss[*s++][*q++];
		if (sum > right_sum) { right_sum = sum; end1 = q; }
	    }
#else
	    right_sum = sum = 0;
	    while (q < seq1_end && sum >= right_sum - X)
		if ((sum += ss[*s++][*q++]) > right_sum) {
			right_sum = sum;
			end1 = q;
		}
#endif
	}

	{ /* extend to the left */
	    const uchar *q = beg1 = (seq1 + pos1) - W;
	    const uchar *s = (seq2+pos2)-W;
	    const int end = diag_lev[diag]-W;
	    const int end_ = (end < 0) ? 0 : end;
	    const int end__ = (end_ < -diag) ? -diag : end_;
	    const uchar *const seq1_end = seq1+end__;
	    int sum;

	    left_sum = sum = 0;
	    while (q > seq1_end  && sum >= left_sum - X)
		if ((sum += ss[*--s][*--q]) > left_sum) {
			left_sum = sum;
			beg1 = q;
		}
	}

	/*
	* Our improper scoring for ambiguous nucleotides could be rectified
	* by scanning the MSP to confirm its score.
	*/
	{   int score;
	    if ((score = ss['A']['A']*W + right_sum + left_sum) >= K)
		msp_add(mt, end1-beg1, beg1-seq1, beg1-seq1+diag, score, 0);
	}

	diag_lev[diag] = (end1 - seq1) + W;

	return 0;
}

msp_t *msp_cons(int len, int pos1, int pos2, int score, int filter)
{
	msp_t *mp = ckalloc(sizeof(*mp));
	mp->len = len;
	mp->pos1 = pos1;
	mp->pos2 = pos2;
	mp->score = score;
	mp->filter = filter;
	return mp;
}

int msp_add(msp_table_t *mt, 
		int len, int pos1, int pos2, int score, int filter)
{
	/*printf("add %d %d %d %d %d\n", len, pos1, pos2, score, filter);*/
	if (mt->num >= mt->size) {
		/*fprintf(stderr, "grow(%d/%d) ", mt->num, mt->size);*/
		mt->size += 15;
		mt->size += mt->size/2;
		mt->msp = ckrealloc(mt->msp, mt->size*sizeof(*mt->msp));
	}
	{ msp_t *const msp = mt->msp + mt->num;
	  msp->len = len;
	  msp->pos1 = pos1;
	  msp->pos2 = pos2;
	  msp->score = score;
	  msp->filter = filter;
	}
	mt->num++;
	/*fprintf(stderr, "(%d/%d)\n", mt->num, mt->size);*/
	return mt->num;
}

static void blast_free_chain(struct blast_table *bt, blast_ecode_t ecode)
{
	blast_epos_node_t **p = pos_getp(bt, ecode);
	blast_epos_node_t *q = *p;

	while (q) {
		blast_epos_node_t *t = q->link;
		ZFREE(q);
		q = t;
	}
	*p = 0; /* clear table entry too */
}

void blast_free_table(struct blast_table *bt)
{
	blast_ecode_t ecode;

        if (bt == 0) return;

	for (ecode = 0; ecode < BLAST_POS_TAB_SIZE; ++ecode)
		blast_free_chain(bt, ecode);
	ZFREE(bt->u.epos_tab);
	ZFREE(bt);
}

void msp_free_table(msp_table_t *mt)
{
	if (mt && mt->msp) ZFREE(mt->msp);
	if (mt) ZFREE(mt);
}
